from .described_estimator import DescribedEstimator


__version__ = '0.1.3'

__title__ = 'destimator'
__description__ = 'A metadata-saving proxy for scikit-learn etimators.'
__uri__ = 'https://github.com/rainforestapp/destimator'

__author__ = 'Maciej Gryka'
__email__ = 'maciej@rainforestqa.com'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2015 RainforestQA'
